/**
 *******************************************************************************
 * @file    usbd_msc.c
 * @brief   MSC(BOT) Function Code for TOSHIBA 'TMPM366' Device Series
 * @version V0.202
 * @date    2014/05/20
 *
 * DO NOT USE THIS SOFTWARE WITHOUT THE SOFTWARE LICENSE AGREEMENT.
 *
 * (C)Copyright TOSHIBA ELECTRONIC DEVICES & STORAGE CORPORATION 2017 All rights reserved
 *******************************************************************************
 */

#include <string.h>
#include "usbd_hw.h"
#include "usbd_device_request.h"
#include "usbd_var.h"
#include "usbd_msc.h"
#include "usbd_trans.h"
#include "usbd_descriptor_msc.h"

/**
  * @brief Callback information
  */
MSC_CB_t gMSC_CB;
/**
  * @brief Don't care status
  */
bool gbStatDCMSC;

/**
  * @brief USB connect status
  */
static bool l_bUsbStat;
/**
  * @brief Max LUN buffer
  */
static uint32_t l_ulMaxLUN_Buffer;
/**
  * @brief Max LUN buffer pointer
  */
static uint8_t *l_pucMaxLUN = (uint8_t *)&l_ulMaxLUN_Buffer;
/**
  * @brief CBW buffer
  */
static uint32_t l_aulCBW_Buffer[USBD_MSC_BOT_CBW_BUFFERSIZE/4U];
/**
  * @brief CBW buffer pointer
  */
static CBW_t *l_ptCbw = (CBW_t *)l_aulCBW_Buffer;
/**
  * @brief CSW buffer
  */
static CSW_t l_tCsw;
/**
  * @brief Reset recovery state
  */
static uint8_t l_ucStatResetRecovery;

/* Prototypes */
static void _USBD_MSC_CB_SendData(USBD_EPx);
static void _USBD_MSC_CB_RecvData(USBD_EPx);
static void _USBD_MSC_RequestCBW(void);
static void _USBD_MSC_CB_RequestCBW(USBD_EPx);
static bool _USBD_MSC_SendCSW(void);
static void _USBD_MSC_CB_SendCSW(USBD_EPx);
static bool _USBD_MSC_MassStorageReset(void);
static bool _USBD_MSC_GetMaxLun(void);
static void _USBD_MSC_StallAllEp(void);
static void _USBD_MSC_XferredCancel(void);

/**
  * @brief Endpoint information
  */
extern EpTransInfo_t gEpTransInfo[CNumEndpointSupport];

/******************************************************************************/

/**
  * @brief  Initialize MSC(BOT) driver
  * @param  p_msc_cb : MSC(BOT) callback information pointer.
  * @retval None
  */
void usbd_msc_init(MSC_CB_t *p_msc_cb)
{
    /* Set of callback functions */
    gMSC_CB.pfnConnStat         = p_msc_cb->pfnConnStat;
    gMSC_CB.pfnMassStorageReset = p_msc_cb->pfnMassStorageReset;
    gMSC_CB.pfnGetMaxLun        = p_msc_cb->pfnGetMaxLun;
    gMSC_CB.pfnRecvCbw          = p_msc_cb->pfnRecvCbw;
    gMSC_CB.pfnSendData         = p_msc_cb->pfnSendData;
    gMSC_CB.pfnRecvData         = p_msc_cb->pfnRecvData;
    gMSC_CB.pfnResume           = p_msc_cb->pfnResume;
    gMSC_CB.pfnSuspend          = p_msc_cb->pfnSuspend;

    /* Initialization of an internal variable */
    l_bUsbStat            = false;
    *l_pucMaxLUN          = (uint8_t)USBD_MSC_BOT_MAX_LUN;
    l_ucStatResetRecovery = USBD_MSC_BOT_RR_NO_REQUEST;

    return;
}

/**
  * @brief  Class request process of MSC(BOT)
  * @param  request : Index of request.
  * @retval bStat : true or false.
  */
bool usbd_interrupt_MSC_req(uint32_t request)
{
    bool bStat;

    if ((f_bmDirection == CbmRequestType_Direction_DtoH)
     && (g_RequestType == CbmRequestType_Type_Class)
     && (g_Recipient == CbmRequestType_Recipient_Interface)) {
        switch (request) {
        case USBD_MSC_BOT_GET_MAX_LUN:
            /* GET_MAX_LUN */
            bStat = _USBD_MSC_GetMaxLun();
            break;
        default:
            /* Not supported request */
            bStat = false;
            break;
        }

    } else if ((f_bmDirection == CbmRequestType_Direction_HtoD)
     && (g_RequestType == CbmRequestType_Type_Class)
     && (g_Recipient == CbmRequestType_Recipient_Interface)) {
        switch( request ) {
        case USBD_MSC_BOT_MASS_STORAGE_RESET:
            /* MASS_STORAGE_RESET */
            bStat = _USBD_MSC_MassStorageReset();
            break;
        default:
            /* Not supported request */
            bStat = false;
            break;
        }
    } else {
        /* Not supported request */
        bStat = false;
    }

    return bStat;
}

/**
  * @brief  Send data on bulk IN endpoint
  * @param  ulSize : Send data size.
  * @param  pucData : Send data buffer pointer.
  * @param  pvInfo : Extended information.
  *    This parameter set operation code of SCSI command in MSC function.
  * @retval bStat : true or false.
  */
bool USBD_MSC_SendData(uint32_t ulSize, uint8_t *pucData, void *pvInfo)
{
    bool    bStat;

    /* Send data on EP1 (Bulk-IN) */
    bStat = usbd_Epx_Send(USBD_EP1, ulSize, pucData, pvInfo, _USBD_MSC_CB_SendData);

    return bStat;
}

/**
  * @brief  Callback for USBD_MSC_SendData
  * @param  epNum : Endpoint number.
  *    This parameter is only USBD_EP1.
  * @retval None
  */
static void _USBD_MSC_CB_SendData(USBD_EPx epNum)
{
    if (gMSC_CB.pfnSendData != NULL) {
        /* Callback */
        (*gMSC_CB.pfnSendData)(gEpTransInfo[epNum].ulTransSize, gEpTransInfo[epNum].pucBuf, gEpTransInfo[epNum].pvInfo, gEpTransInfo[epNum].usEpStat);
    }
}

/**
  * @brief  Receive data on bulk OUT endpoint
  * @param  ulSize : Receive data size.
  * @param  pucData : Receive data buffer pointer.
  * @param  pvInfo : Extended information.
  *    This parameter set operation code of SCSI command in MSC function.
  * @retval bStat : true or false.
  */
bool USBD_MSC_RecvData(uint32_t ulSize, uint8_t *pucData, void *pvInfo)
{
    bool     bStat;

    /* Receive data on EP2 (Bulk-OUT) */
    bStat = usbd_Epx_Recv(USBD_EP2, ulSize, pucData, pvInfo, _USBD_MSC_CB_RecvData);

    return bStat;
}

/**
  * @brief  Callback for USBD_MSC_RecvData
  * @param  epNum : Endpoint number.
  *    This parameter is only USBD_EP2.
  * @retval true
  */
static void _USBD_MSC_CB_RecvData(USBD_EPx epNum)
{
    if (gMSC_CB.pfnRecvData != NULL) {
        /* Callback */
        (*gMSC_CB.pfnRecvData)(gEpTransInfo[epNum].ulTransSize, gEpTransInfo[epNum].pucBuf, gEpTransInfo[epNum].pvInfo, gEpTransInfo[epNum].usEpStat);
    }

    return;
}

/**
  * @brief  Request CBW
  * @param  None
  * @retval None
  */
static void _USBD_MSC_RequestCBW(void)
{
    /* Clear CSW information */
    l_tCsw.ulCSWSignature   = 0U;
    l_tCsw.ulCSWTag         = 0U;
    l_tCsw.ulCSWDataResidue = 0U;
    l_tCsw.ucCSWStatus      = 0U;

    /* Receive CBW on EP2 (Bulk-OUT) */
    gbStatDCMSC = usbd_Epx_Recv(USBD_EP2, USBD_MSC_BOT_CBW_BUFFERSIZE, (uint8_t *)l_ptCbw, NULL, _USBD_MSC_CB_RequestCBW);
    /* Don't care status */

    return;
}

/**
  * @brief  Callback for _USBD_MSC_RequestCBW
  * @param  epNum : Endpoint number.
  *    This parameter is only USBD_EP2.
  * @retval None
  */
static void _USBD_MSC_CB_RequestCBW(USBD_EPx epNum)
{
    CBW_t *ptCbw = (CBW_t *)gEpTransInfo[epNum].pucBuf;

    if (gEpTransInfo[epNum].usEpStat == EPSTAT_COMPLETE) {
        /* Check of CBW */
        if ((USBD_MSC_BOT_CBWLEN  != gEpTransInfo[epNum].ulTransSize)
         || (USBD_MSC_BOT_CBW_SIG != ptCbw->ulCBWSignature)) {
            /* Bulk IN endpoint and bulk OUT endpoint is set to STALL  */
            /* Device holds this state until it receives reset recovery */
            _USBD_MSC_StallAllEp();
            return;
        }

        if (gMSC_CB.pfnRecvCbw != NULL) {
            /* Callback */
            (*gMSC_CB.pfnRecvCbw)(gEpTransInfo[epNum].pucBuf);
        }
    }

    return;
}

/**
  * @brief  Send response (Send CSW)
  * @param  ucStatus : Command status.
  *    This parameter can be one of the following values :
  *      USBD_MSC_BOT_COMMAND_PASSED,
  *      USBD_MSC_BOT_COMMAND_FAILED,
  *      USBD_MSC_BOT_COMMAND_PHASE_ERROR
  * @param  ulXferredLength : Transferred length.
  * @param  bStall : STALL flag.
  *    This parameter can be one of the following values :
  *      USBD_MSC_BOT_REQ_STALL,
  *      USBD_MSC_BOT_NO_STALL
  * @retval bStat : true or false.
  */
bool USBD_MSC_SetResponse(uint8_t ucStatus, uint32_t ulXferredLength, bool bStall)
{
    bool        bStat = true;
    USBD_EPx    epNum;

    /* Create CSW */
    l_tCsw.ulCSWSignature   = USBD_MSC_BOT_CSW_SIG;
    l_tCsw.ulCSWTag         = l_ptCbw->ulCBWTag;
    l_tCsw.ulCSWDataResidue = l_ptCbw->ulCBWDataTransferLength - ulXferredLength;
    l_tCsw.ucCSWStatus      = ucStatus;

    if (l_tCsw.ulCSWDataResidue == 0U) {
        if (bStall == USBD_MSC_BOT_REQ_STALL) {
            /* Set EP1 STALL */
            usbd_set_epx_feature(USBD_EP1, USBD_CMD_EP_STALL);
        } else {
            /* Send CSW */
            bStat = _USBD_MSC_SendCSW();
        }
    } else {
        if (bStall == USBD_MSC_BOT_REQ_STALL) {
            if ((l_ptCbw->ucCBWFlag & USBD_MSC_BOT_CBW_FLAGS_DBIT) == USBD_MSC_BOT_CBW_FLAGS_D2H) {
                epNum = USBD_EP1;
            } else {
                epNum = USBD_EP2;
            }
            /* Set EP1 or EP2 STALL */
            usbd_set_epx_feature(epNum, USBD_CMD_EP_STALL);
        } else {
            /* Send CSW */
            bStat = _USBD_MSC_SendCSW();
        }
    }

    return bStat;
}

/**
  * @brief  Send CSW
  * @param  None
  * @retval bStat : true or false.
  */
static bool _USBD_MSC_SendCSW(void)
{
    bool     bStat;

    /* Send CSW on EP1 (Bulk-IN) */
    bStat = usbd_Epx_Send(USBD_EP1, USBD_MSC_BOT_CSWLEN, (uint8_t *)&l_tCsw, NULL, _USBD_MSC_CB_SendCSW);

    return bStat;
}

/**
  * @brief  Callback for _USBD_MSC_SendCSW
  * @param  epNum : Endpoint number.
  *    This parameter is only USBD_EP1.
  * @retval None
  */
static void _USBD_MSC_CB_SendCSW(USBD_EPx epNum)
{
    if (gEpTransInfo[epNum].usEpStat == EPSTAT_COMPLETE) {
        /* Request CBW */
        _USBD_MSC_RequestCBW();
    }
}

/**
  * @brief  Callback for CLEAR_FEATURE (direct link)
  * @param  epNum : Endpoint number.
  *    This parameter can be one of the following values :
  *      USBD_EP1, USBD_EP2
  * @param  usValue : Request Value (no use).
  * @retval None
  */
void  USBD_MSC_CB_ClearFeature(USBD_EPx epNum, uint16_t usValue)
{
    switch (l_ucStatResetRecovery) {
    case USBD_MSC_BOT_RR_NO_REQUEST:
        if (USBD_MSC_BOT_CSW_SIG == l_tCsw.ulCSWSignature) {
            /* Send CSW */
            _USBD_MSC_SendCSW();
        } else if (USBD_EP2 == epNum) {
            /* Cancellation to a transmission demand */
            _USBD_MSC_XferredCancel();
            /* Request CBW */
            _USBD_MSC_RequestCBW();
        }
        break;

    case USBD_MSC_BOT_RR_REQUESTED:
        /* Set EP2 STALL */
        usbd_set_epx_feature(USBD_EP2, USBD_CMD_EP_STALL);
        /* Set EP1 STALL */
        usbd_set_epx_feature(USBD_EP1, USBD_CMD_EP_STALL);
        break;

    case USBD_MSC_BOT_RR_RCV_MSR:
        if (USBD_EP2 == epNum) {
            /* Set the no request state */
            l_ucStatResetRecovery = USBD_MSC_BOT_RR_NO_REQUEST;
            /* Request CBW */
            _USBD_MSC_RequestCBW();
        }
        break;

    default:
        /* no case */
        break;
    }
}

/**
  * @brief  Received MASS_STORAGE_RESET
  * @param  None
  * @retval bStat : true or false.
  */
static bool _USBD_MSC_MassStorageReset(void)
{
    bool bStat    = false;
    bool bRRState = true;

    if ((g_wValue  == USBD_MSC_BOT_MSR_VALUE)
     && (g_wIndex  == CbInterfaceNumber1_0)
     && (g_wLength == USBD_MSC_BOT_MSR_LENGTH)) {
        /* Cancellation to a transmission demand */
        _USBD_MSC_XferredCancel();

        if (gMSC_CB.pfnMassStorageReset != NULL) {
            /* Callback */
            bRRState = (*gMSC_CB.pfnMassStorageReset)();
        }

        if (bRRState == true) {
            /* Set the received mass storage reset state */
            l_ucStatResetRecovery = USBD_MSC_BOT_RR_RCV_MSR;
        } else {
            /* Set the request reset recovery state */
            l_ucStatResetRecovery = USBD_MSC_BOT_RR_REQUESTED;
        }
        
        bStat = true;
    }
    
    return bStat;
}

/**
  * @brief  Received GET_MAX_LUN
  * @param  None
  * @retval bStat : true or false.
  */
static bool _USBD_MSC_GetMaxLun(void)
{
    bool bStat = false;
    
    if ((g_wValue  == USBD_MSC_BOT_GML_VALUE)
     && (g_wIndex  == CbInterfaceNumber1_0)
     && (g_wLength == USBD_MSC_BOT_GML_LENGTH)) {
        if (gMSC_CB.pfnGetMaxLun != NULL) {
            /* Callback */
            *l_pucMaxLUN = (*gMSC_CB.pfnGetMaxLun)();
        }

        /* Send Max LUN on EP0 (Control-IN) */
        bStat = usbd_Epx_Send(USBD_EP0, USBD_MSC_BOT_GML_LENGTH, l_pucMaxLUN, NULL, NULL);

        if (bStat == false) {
            /* Set EP0 STALL */
            usbd_set_epx_feature(USBD_EP0, USBD_CMD_EP_STALL);
        }
    }
    
    return bStat;
}

/**
  * @brief  STALL all endpoints
  * @param  None
  * @retval None
  */
static void  _USBD_MSC_StallAllEp(void)
{
    /* Set the request reset recovery state */
    l_ucStatResetRecovery = USBD_MSC_BOT_RR_REQUESTED;

    /* Set EP2 STALL */
    usbd_set_epx_feature(USBD_EP2, USBD_CMD_EP_STALL);
    /* Set EP1 STALL */
    usbd_set_epx_feature(USBD_EP1, USBD_CMD_EP_STALL);
}

/**
  * @brief  Cancel transfer
  * @param  None
  * @retval None
  */
static void _USBD_MSC_XferredCancel(void)
{
    /* Cancellation to a EP2 (Bulk-OUT) transmission demand */
    usbd_Epx_Abort(USBD_EP2);
    /* Cancellation to a EP1 (Bulk-IN) transmission demand */
    usbd_Epx_Abort(USBD_EP1);
}

/**
  * @brief  USB connect (Call at CONFIGURED state)
  * @param  None
  * @retval None
  */
void USBD_MSC_Connect(void)
{
    /* Check local USB state */
    if (l_bUsbStat == false) {
        l_bUsbStat = true;
    
        /* Initialize reset recovery status */
        l_ucStatResetRecovery = USBD_MSC_BOT_RR_NO_REQUEST;

        if (gMSC_CB.pfnConnStat != NULL) {
            /* Callback */
            (*gMSC_CB.pfnConnStat)(l_bUsbStat);
        }

        /* Request CBW */
        _USBD_MSC_RequestCBW();
    }
}

/**
  * @brief  USB disconnect (Call at Non-CONFIGURED State)
  * @param  None
  * @retval None
  */
void USBD_MSC_DisConnect(void)
{
    /* Check local USB state */
    if (l_bUsbStat == true) {
        l_bUsbStat = false;

        if (gMSC_CB.pfnConnStat != NULL) {
            /* Callback */
            (*gMSC_CB.pfnConnStat)(l_bUsbStat);
        }

        /* Cancellation to a transmission demand */
        _USBD_MSC_XferredCancel();
    }
}

/* EOC */
